@@ -10,8 +10,11 @@ import base64
 import json
 import math
 import os
+import re
 import struct
+import typing
 import zlib
+from typing import Any, Callable, Union
 
 from jinja2 import Environment, PackageLoader
 
@@ -25,13 +28,16 @@ try:
 except ImportError:
     np = None
 
+if typing.TYPE_CHECKING:
+    from branca.colormap import ColorMap
+
 
 rootpath = os.path.abspath(os.path.dirname(__file__))
 
 
 def get_templates():
     """Get Jinja templates."""
-    return Environment(loader=PackageLoader('branca', 'templates'))
+    return Environment(loader=PackageLoader("branca", "templates"))
 
 
 def legend_scaler(legend_values, max_labels=10.0):
@@ -44,11 +50,11 @@ def legend_scaler(legend_values, max_labels=10.0):
     if len(legend_values) < max_labels:
         legend_ticks = legend_values
     else:
-        spacer = int(math.ceil(len(legend_values)/max_labels))
+        spacer = int(math.ceil(len(legend_values) / max_labels))
         legend_ticks = []
         for i in legend_values[::spacer]:
             legend_ticks += [i]
-            legend_ticks += ['']*(spacer-1)
+            legend_ticks += [""] * (spacer - 1)
     return legend_ticks
 
 
@@ -63,6 +69,7 @@ def linear_gradient(hexList, nColors):
     >>> linear_gradient([(0, 0, 0), (255, 0, 0), (255, 255, 0)], 100)
 
     """
+
     def _scale(start, finish, length, i):
         """
         Return the value correct value of a number that is in between start
@@ -73,22 +80,22 @@ def linear_gradient(hexList, nColors):
 
         fraction = float(i) / (length - 1)
         raynge = int(finish, base) - int(start, base)
-        thex = hex(int(int(start, base) + fraction * raynge)).split('x')[-1]
+        thex = hex(int(int(start, base) + fraction * raynge)).split("x")[-1]
         if len(thex) != 2:
-            thex = '0' + thex
+            thex = "0" + thex
         return thex
 
     allColors = []
     # Separate (R, G, B) pairs.
     for start, end in zip(hexList[:-1], hexList[1:]):
-        # Linearly intepolate between pair of hex ###### values and
+        # Linearly interpolate between pair of hex ###### values and
         # add to list.
         nInterpolate = 765
         for index in range(nInterpolate):
             r = _scale(start[1:3], end[1:3], nInterpolate, index)
             g = _scale(start[3:5], end[3:5], nInterpolate, index)
             b = _scale(start[5:7], end[5:7], nInterpolate, index)
-            allColors.append(''.join(['#', r, g, b]))
+            allColors.append("".join(["#", r, g, b]))
 
     # Pick only nColors colors from the total list.
     result = []
@@ -109,39 +116,42 @@ def color_brewer(color_code, n=6):
     minimum_n = 3
 
     if not isinstance(n, int):
-        raise TypeError('n has to be an int, not a %s' % type(n))
+        raise TypeError("n has to be an int, not a %s" % type(n))
 
     # Raise an error if the n requested is greater than the maximum.
     if n > maximum_n:
-        raise ValueError('The maximum number of colors in a'
-                         ' ColorBrewer sequential color series is 253')
+        raise ValueError(
+            "The maximum number of colors in a"
+            " ColorBrewer sequential color series is 253",
+        )
     if n < minimum_n:
-        raise ValueError('The minimum number of colors in a'
-                         ' ColorBrewer sequential color series is 3')
+        raise ValueError(
+            "The minimum number of colors in a"
+            " ColorBrewer sequential color series is 3",
+        )
 
     if not isinstance(color_code, str):
-        raise ValueError('color should be a string, not a {}.'
-                         .format(type(color_code)))
-    if color_code[-2:] == '_r':
+        raise ValueError(f"color should be a string, not a {type(color_code)}.")
+    if color_code[-2:] == "_r":
         base_code = color_code[:-2]
-        core_color_code = base_code + '_' + str(n).zfill(2)
+        core_color_code = base_code + "_" + str(n).zfill(2)
         color_reverse = True
     else:
         base_code = color_code
-        core_color_code = base_code + '_' + str(n).zfill(2)
+        core_color_code = base_code + "_" + str(n).zfill(2)
         color_reverse = False
 
-    with open(os.path.join(rootpath, '_schemes.json')) as f:
+    with open(os.path.join(rootpath, "_schemes.json")) as f:
         schemes = json.loads(f.read())
 
-    with open(os.path.join(rootpath, 'scheme_info.json')) as f:
+    with open(os.path.join(rootpath, "scheme_info.json")) as f:
         scheme_info = json.loads(f.read())
 
-    with open(os.path.join(rootpath, 'scheme_base_codes.json')) as f:
-        core_schemes = json.loads(f.read())['codes']
+    with open(os.path.join(rootpath, "scheme_base_codes.json")) as f:
+        core_schemes = json.loads(f.read())["codes"]
 
     if base_code not in core_schemes:
-        raise ValueError(base_code + ' is not a valid ColorBrewer code')
+        raise ValueError(base_code + " is not a valid ColorBrewer code")
 
     explicit_scheme = True
     if schemes.get(core_color_code) is None:
@@ -150,34 +160,41 @@ def color_brewer(color_code, n=6):
     # Only if n is greater than the scheme length do we interpolate values.
     if not explicit_scheme:
         # Check to make sure that it is not a qualitative scheme.
-        if scheme_info[base_code] == 'Qualitative':
+        if scheme_info[base_code] == "Qualitative":
             matching_quals = []
             for key in schemes:
-                if base_code + '_' in key:
-                    matching_quals.append(int(key.split('_')[1]))
+                if base_code + "_" in key:
+                    matching_quals.append(int(key.split("_")[1]))
 
-            raise ValueError('Expanded color support is not available'
-                             ' for Qualitative schemes; restrict the'
-                             ' number of colors for the ' + base_code +
-                             ' code to between ' + str(min(matching_quals)) +
-                             ' and ' + str(max(matching_quals))
-                             )
+            raise ValueError(
+                "Expanded color support is not available"
+                " for Qualitative schemes; restrict the"
+                " number of colors for the "
+                + base_code
+                + " code to between "
+                + str(min(matching_quals))
+                + " and "
+                + str(max(matching_quals)),
+            )
         else:
             longest_scheme_name = base_code
             longest_scheme_n = 0
             for sn_name in schemes.keys():
-                if '_' not in sn_name:
+                if "_" not in sn_name:
                     continue
-                if sn_name.split('_')[0] != base_code:
+                if sn_name.split("_")[0] != base_code:
                     continue
-                if int(sn_name.split('_')[1]) > longest_scheme_n:
+                if int(sn_name.split("_")[1]) > longest_scheme_n:
                     longest_scheme_name = sn_name
-                    longest_scheme_n = int(sn_name.split('_')[1])
+                    longest_scheme_n = int(sn_name.split("_")[1])
 
             if not color_reverse:
                 color_scheme = linear_gradient(schemes.get(longest_scheme_name), n)
             else:
-                color_scheme = linear_gradient(schemes.get(longest_scheme_name)[::-1], n)
+                color_scheme = linear_gradient(
+                    schemes.get(longest_scheme_name)[::-1],
+                    n,
+                )
     else:
         if not color_reverse:
             color_scheme = schemes.get(core_color_code, None)
@@ -202,16 +219,14 @@ def split_six(series=None):
 
     """
     if pd is None:
-        raise ImportError('The Pandas package is required'
-                          ' for this functionality')
+        raise ImportError("The Pandas package is required" " for this functionality")
     if np is None:
-        raise ImportError('The NumPy package is required'
-                          ' for this functionality')
+        raise ImportError("The NumPy package is required" " for this functionality")
 
     def base(x):
         if x > 0:
             base = pow(10, math.floor(math.log10(x)))
-            return round(x/base)*base
+            return round(x / base) * base
         else:
             return 0
 
@@ -221,7 +236,7 @@ def split_six(series=None):
     return [base(np.percentile(arr, x)) for x in quants]
 
 
-def image_to_url(image, colormap=None, origin='upper'):
+def image_to_url(image, colormap=None, origin="upper"):
     """Infers the type of an image argument and transforms it into a URL.
 
     Parameters
@@ -241,35 +256,43 @@ def image_to_url(image, colormap=None, origin='upper'):
         It must output iterables of length 3 or 4, with values between
         0. and 1.  Hint : you can use colormaps from `matplotlib.cm`.
     """
-    if hasattr(image, 'read'):
+    if hasattr(image, "read"):
         # We got an image file.
-        if hasattr(image, 'name'):
+        if hasattr(image, "name"):
             # We try to get the image format from the file name.
-            fileformat = image.name.lower().split('.')[-1]
+            fileformat = image.name.lower().split(".")[-1]
         else:
-            fileformat = 'png'
-        url = 'data:image/{};base64,{}'.format(
-            fileformat, base64.b64encode(image.read()).decode('utf-8'))
-    elif (not (isinstance(image, str) or
-               isinstance(image, bytes))) and hasattr(image, '__iter__'):
+            fileformat = "png"
+        url = "data:image/{};base64,{}".format(
+            fileformat,
+            base64.b64encode(image.read()).decode("utf-8"),
+        )
+    elif (not (isinstance(image, str) or isinstance(image, bytes))) and hasattr(
+        image,
+        "__iter__",
+    ):
         # We got an array-like object.
         png = write_png(image, origin=origin, colormap=colormap)
-        url = 'data:image/png;base64,' + base64.b64encode(png).decode('utf-8')
+        url = "data:image/png;base64," + base64.b64encode(png).decode("utf-8")
     else:
         # We got an URL.
         url = json.loads(json.dumps(image))
 
-    return url.replace('\n', ' ')
+    return url.replace("\n", " ")
 
 
-def write_png(data, origin='upper', colormap=None):
+def write_png(
+    data: Any,
+    origin: str = "upper",
+    colormap: Union["ColorMap", Callable, None] = None,
+) -> bytes:
     """
     Transform an array of data into a PNG string.
     This can be written to disk using binary I/O, or encoded using base64
     for an inline PNG like this:
 
     >>> png_str = write_png(array)
-    >>> 'data:image/png;base64,'+png_str.encode('base64')
+    >>> "data:image/png;base64," + png_str.encode("base64")
 
     Inspired from
     http://stackoverflow.com/questions/902761/saving-a-numpy-array-as-an-image
@@ -278,43 +301,46 @@ def write_png(data, origin='upper', colormap=None):
     ----------
     data: numpy array or equivalent list-like object.
          Must be NxM (mono), NxMx3 (RGB) or NxMx4 (RGBA)
-
     origin : ['upper' | 'lower'], optional, default 'upper'
         Place the [0,0] index of the array in the upper left or lower left
         corner of the axes.
-
-    colormap : callable, used only for `mono` image.
-        Function of the form [x -> (r,g,b)] or [x -> (r,g,b,a)]
-        for transforming a mono image into RGB.
-        It must output iterables of length 3 or 4, with values between
-        0. and 1.  Hint: you can use colormaps from `matplotlib.cm`.
+    colormap : ColorMap subclass or callable, optional
+        Only needed to transform mono images into RGB. You have three options:
+        - use a subclass of `ColorMap` like `LinearColorMap`
+        - use a colormap from `matplotlib.cm`
+        - use a custom function of the form [x -> (r,g,b)] or [x -> (r,g,b,a)].
+          It must output iterables of length 3 or 4 with values between 0 and 1.
 
     Returns
     -------
     PNG formatted byte string
     """
-    if np is None:
-        raise ImportError('The NumPy package is required'
-                          ' for this functionality')
+    from branca.colormap import ColorMap
 
-    if colormap is None:
-        def colormap(x):
-            return (x, x, x, 1)
+    if np is None:
+        raise ImportError("The NumPy package is required" " for this functionality")
+
+    if isinstance(colormap, ColorMap):
+        colormap_callable = colormap.rgba_floats_tuple
+    elif callable(colormap):
+        colormap_callable = colormap
+    else:
+        colormap_callable = lambda x: (x, x, x, 1)  # noqa E731
 
     array = np.atleast_3d(data)
     height, width, nblayers = array.shape
 
     if nblayers not in [1, 3, 4]:
-        raise ValueError('Data must be NxM (mono), '
-                         'NxMx3 (RGB), or NxMx4 (RGBA)')
+        raise ValueError("Data must be NxM (mono), " "NxMx3 (RGB), or NxMx4 (RGBA)")
     assert array.shape == (height, width, nblayers)
 
     if nblayers == 1:
-        array = np.array(list(map(colormap, array.ravel())))
+        array = np.array(list(map(colormap_callable, array.ravel())))
         nblayers = array.shape[1]
         if nblayers not in [3, 4]:
-            raise ValueError('colormap must provide colors of'
-                             'length 3 (RGB) or 4 (RGBA)')
+            raise ValueError(
+                "colormap must provide colors of" "length 3 (RGB) or 4 (RGBA)",
+            )
         array = array.reshape((height, width, nblayers))
     assert array.shape == (height, width, nblayers)
 
@@ -325,51 +351,88 @@ def write_png(data, origin='upper', colormap=None):
     assert nblayers == 4
 
     # Normalize to uint8 if it isn't already.
-    if array.dtype != 'uint8':
-        array = array * 255./array.max(axis=(0, 1)).reshape((1, 1, 4))
-        array = array.astype('uint8')
+    if array.dtype != "uint8":
+        with np.errstate(divide="ignore", invalid="ignore"):
+            array = array * 255.0 / array.max(axis=(0, 1)).reshape((1, 1, 4))
+            array[~np.isfinite(array)] = 0
+        array = array.astype("uint8")
 
     # Eventually flip the image.
-    if origin == 'lower':
+    if origin == "lower":
         array = array[::-1, :, :]
 
     # Transform the array to bytes.
-    raw_data = b''.join([b'\x00' + array[i, :, :].tobytes()
-                         for i in range(height)])
+    raw_data = b"".join([b"\x00" + array[i, :, :].tobytes() for i in range(height)])
 
     def png_pack(png_tag, data):
         chunk_head = png_tag + data
-        return (struct.pack('!I', len(data)) +
-                chunk_head +
-                struct.pack('!I', 0xFFFFFFFF & zlib.crc32(chunk_head)))
+        return (
+            struct.pack("!I", len(data))
+            + chunk_head
+            + struct.pack("!I", 0xFFFFFFFF & zlib.crc32(chunk_head))
+        )
 
-    return b''.join([
-        b'\x89PNG\r\n\x1a\n',
-        png_pack(b'IHDR', struct.pack('!2I5B', width, height, 8, 6, 0, 0, 0)),
-        png_pack(b'IDAT', zlib.compress(raw_data, 9)),
-        png_pack(b'IEND', b'')])
+    return b"".join(
+        [
+            b"\x89PNG\r\n\x1a\n",
+            png_pack(b"IHDR", struct.pack("!2I5B", width, height, 8, 6, 0, 0, 0)),
+            png_pack(b"IDAT", zlib.compress(raw_data, 9)),
+            png_pack(b"IEND", b""),
+        ],
+    )
 
 
 def _camelify(out):
-    return (''.join(['_' + x.lower() if i < len(out)-1 and x.isupper() and out[i+1].islower()  # noqa
-         else x.lower() + '_' if i < len(out)-1 and x.islower() and out[i+1].isupper()  # noqa
-         else x.lower() for i, x in enumerate(list(out))])).lstrip('_').replace('__', '_')  # noqa
+    return (
+        (
+            "".join(
+                [
+                    (
+                        "_" + x.lower()
+                        if i < len(out) - 1
+                        and x.isupper()
+                        and out[i + 1].islower()  # noqa
+                        else (
+                            x.lower() + "_"
+                            if i < len(out) - 1
+                            and x.islower()
+                            and out[i + 1].isupper()  # noqa
+                            else x.lower()
+                        )
+                    )
+                    for i, x in enumerate(list(out))
+                ],
+            )
+        )
+        .lstrip("_")
+        .replace("__", "_")
+    )  # noqa
 
 
 def _parse_size(value):
-    try:
-        if isinstance(value, int) or isinstance(value, float):
-            value_type = 'px'
-            value = float(value)
-            assert value > 0
+    if isinstance(value, (int, float)):
+        return float(value), "px"
+    elif isinstance(value, str):
+        # match digits or a point, possibly followed by a space,
+        # followed by a unit: either 1 to 5 letters or a percent sign
+        match = re.fullmatch(r"([\d.]+)\s?(\w{1,5}|%)", value.strip())
+        if match:
+            return float(match.group(1)), match.group(2)
         else:
-            value_type = '%'
-            value = float(value.strip('%'))
-            assert 0 <= value <= 100
-    except Exception:
-        msg = 'Cannot parse value {!r} as {!r}'.format
-        raise ValueError(msg(value, value_type))
-    return value, value_type
+            raise ValueError(
+                f"Cannot parse {value!r}, it should be a number followed by a unit.",
+            )
+    elif (
+        isinstance(value, tuple)
+        and isinstance(value[0], (int, float))
+        and isinstance(value[1], str)
+    ):
+        # value had been already parsed
+        return (float(value[0]), value[1])
+    else:
+        raise TypeError(
+            f"Cannot parse {value!r}, it should be a number or a string containing a number and a unit.",
+        )
 
 
 def _locations_mirror(x):
@@ -379,8 +442,8 @@ def _locations_mirror(x):
     [[[2, 1], [4, 3]], [6, 5], [8, 7]]
 
     """
-    if hasattr(x, '__iter__'):
-        if hasattr(x[0], '__iter__'):
+    if hasattr(x, "__iter__"):
+        if hasattr(x[0], "__iter__"):
             return list(map(_locations_mirror, x))
         else:
             return list(x[::-1])
@@ -389,9 +452,8 @@ def _locations_mirror(x):
 
 
 def _locations_tolist(x):
-    """Transforms recursively a list of iterables into a list of list.
-    """
-    if hasattr(x, '__iter__'):
+    """Transforms recursively a list of iterables into a list of list."""
+    if hasattr(x, "__iter__"):
         return list(map(_locations_tolist, x))
     else:
         return x
@@ -431,4 +493,4 @@ def iter_points(x):
         else:
             return []
     else:
-        raise ValueError('List/tuple type expected. Got {!r}.'.format(x))
+        raise ValueError(f"List/tuple type expected. Got {x!r}.")
